@@ -1,19 +1,33 @@
+import functools
+
 import logging
+import pika
 
+from pika.frame import Method
+from pika.channel import Channel
+from pika.spec import BasicProperties
+from pika.spec import Basic
+
+from typing import Union, Callable, Set, Dict
 from abc import ABC, abstractmethod
-
 from threading import Thread, Timer
 
-import pika
 from pika import SelectConnection
 from pika.exceptions import (
     ConnectionClosedByBroker,
     StreamLostError,
-    ConnectionWrongStateError
+    ConnectionWrongStateError,
+    ChannelClosedByBroker
 )
 
-from rabbitmq_client.defs import DEFAULT_EXCHANGE, PublishParams
-
+from rabbitmq_client.defs import (
+    PublishParams,
+    QueueParams,
+    ExchangeParams,
+    ConsumeParams,
+    QueueBindParams,
+    MandatoryError
+)
 
 LOGGER = logging.getLogger(__name__)
 
@@ -22,8 +36,52 @@ RECONNECT_REASONS = (
     StreamLostError  # Ungraceful shutdown
 )
 
+# 405 = Resource locked
+# 406 = PRECONDITION FAILED
+PERMANENT_CLOSURE_CODES = (
+    405, 406
+)
+
+
+class DeclarationError:
+
+    def __init__(self, message: str):
+        self.message = message
+
 
 class RMQConnection(ABC):
+
+    class Cache:
+
+        def __init__(self):
+            self.queues: Set[str] = set()
+            self.exchanges: Set[str] = set()
+            self.consumer_tags: Dict[str, str] = dict()
+
+        def store(self, params: Union[QueueParams, ExchangeParams]):
+            if isinstance(params, QueueParams):
+                self.queues.add(params.queue)
+            elif isinstance(params, ExchangeParams):
+                self.exchanges.add(params.exchange)
+
+        def is_cached(self, params: Union[QueueParams, ExchangeParams]
+                      ) -> bool:
+            if isinstance(params, QueueParams):
+                return params.queue in self.queues
+            elif isinstance(params, ExchangeParams):
+                return params.exchange in self.exchanges
+
+        def link_consumer_tag(self, queue, consumer_tag):
+            self.consumer_tags[queue] = consumer_tag
+
+        def consumer_tag(self, consume_params: ConsumeParams) -> str:
+            return self.consumer_tags.get(consume_params.queue)
+
+        def clear(self):
+            self.queues = set()
+            self.exchanges = set()
+            self.consumer_tags = dict()
+
     """
     Abstract base class to be implemented.
 
@@ -53,6 +111,8 @@ class RMQConnection(ABC):
         self._closing = False
         self._restarting = False
 
+        self.cache = RMQConnection.Cache()
+
     @abstractmethod
     def on_ready(self):
         """
@@ -76,7 +136,7 @@ class RMQConnection(ABC):
         pass
 
     @abstractmethod
-    def on_error(self):
+    def on_error(self, error: Union[MandatoryError, DeclarationError]):
         """
         Implementers can handle what should happen when the RMQConnection has
         run into an error, for example an Exchange/Queue declaration failure.
@@ -110,7 +170,7 @@ class RMQConnection(ABC):
             self._connection.close()
             self._restarting = True  # Only if connection could be closed.
         except ConnectionWrongStateError:
-            LOGGER.info("connection closed and could not be restarted")
+            LOGGER.error("connection closed and could not be restarted")
 
     def stop(self):
         """
@@ -128,49 +188,66 @@ class RMQConnection(ABC):
                 LOGGER.info("connection already closed")
 
     def declare_queue(self,
-                      queue_params,
-                      callback=None):
-        """
-        :param queue_params: rabbitmq_client.QueueParams
-        :param callback: callable
-        """
-        LOGGER.info(f"declaring queue: {queue_params.queue}")
+                      queue_params: QueueParams,
+                      callback: Callable):
+        if self.cache.is_cached(queue_params):
+            LOGGER.info(f"queue '{queue_params.queue}' already declared")
+            callback(queue_params.queue)
+        else:
+            LOGGER.info(f"declaring queue '{queue_params.queue}'")
+            cb = functools.partial(self.on_queue_declared,
+                                   queue_params,
+                                   callback)
+            self._channel.queue_declare(
+                queue_params.queue,
+                durable=queue_params.durable,
+                exclusive=queue_params.exclusive,
+                auto_delete=queue_params.auto_delete,
+                arguments=queue_params.arguments,
+                callback=cb
+            )
 
-        self._channel.queue_declare(
-            queue_params.queue,
-            durable=queue_params.durable,
-            exclusive=queue_params.exclusive,
-            auto_delete=queue_params.auto_delete,
-            arguments=queue_params.arguments,
-            callback=callback
-        )
+    def on_queue_declared(self,
+                          queue_params: QueueParams,
+                          callback: Callable,
+                          method_frame: Method):
+        LOGGER.info(f"queue '{method_frame.method.queue}' declared")
+        self.cache.store(queue_params)
+        callback(method_frame.method.queue)
 
     def declare_exchange(self,
-                         exchange_params,
-                         callback=None):
-        """
-        :param exchange_params: rabbitmq_client.ExchangeParams
-        :param callback: callable
-        """
-        LOGGER.info(f"declaring exchange: {exchange_params.exchange}")
+                         exchange_params: ExchangeParams,
+                         callback: Callable):
+        if self.cache.is_cached(exchange_params):
+            LOGGER.info(f"exchange '{exchange_params.exchange}' already "
+                        f"declared")
+            callback()
+        else:
+            LOGGER.info(f"declaring exchange '{exchange_params.exchange}'")
+            cb = functools.partial(self.on_exchange_declared,
+                                   exchange_params,
+                                   callback)
+            self._channel.exchange_declare(
+                exchange_params.exchange,
+                exchange_type=exchange_params.exchange_type,
+                durable=exchange_params.durable,
+                auto_delete=exchange_params.auto_delete,
+                internal=exchange_params.internal,
+                arguments=exchange_params.arguments,
+                callback=cb
+            )
 
-        self._channel.exchange_declare(
-            exchange_params.exchange,
-            exchange_type=exchange_params.exchange_type,
-            durable=exchange_params.durable,
-            auto_delete=exchange_params.auto_delete,
-            internal=exchange_params.internal,
-            arguments=exchange_params.arguments,
-            callback=callback
-        )
+    def on_exchange_declared(self,
+                             exchange_params: ExchangeParams,
+                             callback: Callable,
+                             _method_frame: Method):
+        LOGGER.info(f"exchange '{exchange_params.exchange}' declared")
+        self.cache.store(exchange_params)
+        callback()
 
     def bind_queue(self,
-                   queue_bind_params,
-                   callback=None):
-        """
-        :param queue_bind_params: rabbitmq_client.QueueBindParams
-        :param callback: callable
-        """
+                   queue_bind_params: QueueBindParams,
+                   callback: Callable):
         LOGGER.info(f"binding queue {queue_bind_params.queue} to exchange "
                     f"{queue_bind_params.exchange} with routing key: "
                     f"{queue_bind_params.routing_key}")
@@ -185,38 +262,44 @@ class RMQConnection(ABC):
 
     def basic_consume(self,
                       consume_params,
-                      on_message_callback_override=None,
-                      callback=None):
-        """
-        :param consume_params: rabbitmq_client.ConsumeParams
-        :param on_message_callback_override: callable
-        :param callback: callable
-        """
-        LOGGER.info(f"basic consumer starting for queue: "
-                    f"{consume_params.queue}")
+                      on_message_callback: Callable,
+                      callback: Callable):
+        consumer_tag = self.cache.consumer_tag(consume_params)
+        if consumer_tag is not None:
+            LOGGER.info(f"queue is already consumed from, consumer tag: "
+                        f"{consumer_tag}")
+            callback(consumer_tag)
+        else:
+            LOGGER.info(f"consume starting for queue "
+                        f"'{consume_params.queue}'")
+            cb = functools.partial(self.on_consume_ok,
+                                   consume_params,
+                                   callback)
+            self._channel.basic_consume(
+                consume_params.queue,
+                on_message_callback,
+                auto_ack=consume_params.auto_ack,
+                exclusive=consume_params.exclusive,
+                consumer_tag=consume_params.consumer_tag,
+                arguments=consume_params.arguments,
+                callback=cb
+            )
 
-        self._channel.basic_consume(
-            consume_params.queue,
-            (on_message_callback_override if on_message_callback_override
-             else consume_params.on_message_callback),
-            auto_ack=consume_params.auto_ack,
-            exclusive=consume_params.exclusive,
-            consumer_tag=consume_params.consumer_tag,
-            arguments=consume_params.arguments,
-            callback=callback
-        )
+    def on_consume_ok(self,
+                      consume_params: ConsumeParams,
+                      callback: Callable,
+                      method_frame: Method):
+        LOGGER.info(f"consume started for queue '{consume_params.queue}' "
+                    f"consumer tag: {method_frame.method.consumer_tag}")
+        self.cache.link_consumer_tag(consume_params.queue,
+                                     method_frame.method.consumer_tag)
+        callback(method_frame.method.consumer_tag)
 
     def basic_publish(self,
-                      body,
-                      exchange=DEFAULT_EXCHANGE,
-                      routing_key="",
-                      publish_params=None):
-        """
-        :param body: bytes
-        :param exchange: str
-        :param routing_key: str
-        :param publish_params: rabbitmq_client.PublishParams
-        """
+                      body: bytes,
+                      exchange: str,
+                      routing_key: str,
+                      publish_params: PublishParams = None):
         LOGGER.debug(f"publishing to exchange: {exchange} and routing key: "
                      f"{routing_key}")
 
@@ -236,9 +319,21 @@ class RMQConnection(ABC):
         """
         LOGGER.info("activating confirm delivery mode")
 
+        self._channel.add_on_return_callback(self.on_return)
         self._channel.confirm_delivery(on_delivery_confirmed,
                                        callback=callback)
 
+    def on_return(self,
+                  _channel: Channel,
+                  method_frame: Basic.Return,
+                  _properties: BasicProperties,
+                  _body: bytes):
+        LOGGER.info(f"a message was returned for "
+                    f"exchange '{method_frame.exchange}' and "
+                    f"routing key '{method_frame.routing_key}'")
+        self.on_error(MandatoryError(method_frame.exchange,
+                                     method_frame.routing_key))
+
     def _connect(self):
         """
         Assigns a new pika.SelectConnection to the connection and starts its
@@ -298,16 +393,24 @@ class RMQConnection(ABC):
         :param _connection: pika.SelectConnection
         :param reason: pika.exceptions.?
         """
+        self.cache.clear()
+
         permanent = False
 
         # This should ensure the current thread runs to completion after
         # on_connection_closed handling is done.
         self._connection.ioloop.stop()
 
-        if not self._closing and type(reason) in RECONNECT_REASONS:
+        # Not something user-triggered, but good enough reason to re-connect
+        if self._closing:
+            LOGGER.debug("connection closed")
+            permanent = True
+
+        elif type(reason) in RECONNECT_REASONS:
             LOGGER.debug(f"connection closed: {reason}, attempting reconnect")
             self._reconnect()
 
+        # We're restarting, gogo and restart
         elif self._restarting:
             LOGGER.debug("connection closed, restarting now")
             # Reset to prevent reconnect for any reason
@@ -316,6 +419,7 @@ class RMQConnection(ABC):
                                              daemon=True)
             self._connection_thread.start()
 
+        # something unknown has happened
         else:
             permanent = True
             LOGGER.warning(f"connection closed: {reason}, will not reconnect")
@@ -343,7 +447,6 @@ class RMQConnection(ABC):
                                          daemon=True)
         if self._reconnect_delay == 0:  # retry immediately the first time
             self._connection_thread.start()
-
         else:  # calculate backoff timer
             timer = Timer(
                 self._reconnect_delay,
@@ -381,20 +484,24 @@ class RMQConnection(ABC):
         :param _channel: pika.channel.Channel
         :param reason: pika.exceptions.?
         """
-        LOGGER.warning("channel closed")
+        self.cache.clear()
 
         permanent = False
+        if self._closing:
+            LOGGER.info("a client requested shutdown has closed the channel")
+            permanent = True
 
-        try:
-            # Reply code 406 = PRECONDITION FAILED
-            permanent = True if reason.reply_code == 406 else False
-        except AttributeError:  # Not all reasons have a reply code
-            pass
+        elif (isinstance(reason, ChannelClosedByBroker) and
+              reason.reply_code in PERMANENT_CLOSURE_CODES):
+            LOGGER.critical(f"permanent channel closure due to: {reason}")
+            permanent = True
 
-        # Signal subclass that connection is down.
-        self.on_close(permanent=permanent)
-
-        if permanent:
-            LOGGER.critical(f"connection stopping due to permanent channel "
-                            f"closure: {reason}")
+            # TODO: Declare errors must have further purpose to be useful
+            #  otherwise why are we reporting them?
+            self.on_error(DeclarationError(reason.reply_text))
             self.stop()
+
+        else:
+            LOGGER.warning(f"channel was closed due to: {reason}")
+
+        self.on_close(permanent=permanent)
