@@ -2,15 +2,23 @@ import functools
 import logging
 import uuid
 
+from typing import Callable, Union
 from threading import Lock
 
 from pika.spec import Basic
 
-from rabbitmq_client.connection import RMQConnection
+from rabbitmq_client.connection import (
+    RMQConnection,
+    MandatoryError,
+    DeclarationError
+)
 from rabbitmq_client.defs import (
     ConfirmModeOK,
     DeliveryError,
-    DEFAULT_EXCHANGE
+    DEFAULT_EXCHANGE,
+    QueueParams,
+    ExchangeParams,
+    PublishParams
 )
 
 LOGGER = logging.getLogger(__name__)
@@ -57,6 +65,7 @@ class RMQProducer(RMQConnection):
 
         self._buffered_messages = list()
 
+        # Confirm mode
         self._confirm_mode_active = False
         self._confirm_delivery_callback = None
         self._next_delivery_tag = 1
@@ -99,11 +108,11 @@ class RMQProducer(RMQConnection):
         super().stop()
 
     def publish(self,
-                body,
-                exchange_params=None,
-                routing_key="",
-                queue_params=None,
-                publish_params=None):
+                body: bytes,
+                publish_params: PublishParams = None,
+                queue_params: QueueParams = None,
+                exchange_params: ExchangeParams = None,
+                routing_key: str = "") -> Union[None, str]:
         """
         Publish 'body' towards an exchange (optional routing key) or a queue.
         You must provide EITHER exchange_params OR queue_params to handle
@@ -122,13 +131,6 @@ class RMQProducer(RMQConnection):
         been sent successfully by waiting for their key to be referred to in
         a call to their confirm mode notify callback where the input parameter
         == the generated publish key.
-
-        :param body: bytes
-        :param exchange_params: rabbitmq_client.ExchangeParams
-        :param routing_key: str
-        :param queue_params: rabbitmq_client.QueueParams
-        :param publish_params: rabbitmq_client.PublishParams
-        :returns: str
         """
         # More detailed logging done on connection level
         LOGGER.info("publishing")
@@ -155,12 +157,11 @@ class RMQProducer(RMQConnection):
                 )
         ):
             self._handle_publish(body,
-                                 exchange_params,
-                                 routing_key,
-                                 queue_params,
-                                 publish_params,
-                                 publish_key)
-
+                                 publish_params=publish_params,
+                                 queue_params=queue_params,
+                                 exchange_params=exchange_params,
+                                 routing_key=routing_key,
+                                 publish_key=publish_key)
         else:
             self._buffered_messages.append(
                 RMQPublish(body,
@@ -173,86 +174,55 @@ class RMQProducer(RMQConnection):
 
         return publish_key
 
-    def activate_confirm_mode(self, notify_callback):
-        """
-        :param notify_callback: callable
-        """
+    def activate_confirm_mode(self,
+                              notify_callback: Callable[
+                                  [Union[str, ConfirmModeOK, DeliveryError]],
+                                  None
+                              ]):
         if self.ready and self._confirm_delivery_callback is None:
-            LOGGER.info("activating confirm delivery mode")
             self.confirm_delivery(self.on_delivery_confirmed,
                                   callback=self.on_confirm_select_ok)
 
         self._confirm_delivery_callback = notify_callback
 
     def _handle_publish(self,
-                        body,
-                        exchange_params,
-                        routing_key,
-                        queue_params,
-                        publish_params,
-                        publish_key):
-        """
-        :param body: bytes
-        :param exchange_params: rabbitmq_client.ExchangeParams
-        :param routing_key: str
-        :param queue_params: rabbitmq_client.QueueParams
-        :param publish_params: rabbitmq_client.PublishParams
-        :param publish_key: str
-        """
+                        body: bytes,
+                        publish_params: PublishParams = None,
+                        queue_params: QueueParams = None,
+                        exchange_params: ExchangeParams = None,
+                        routing_key: str = "",
+                        publish_key: Union[None, str] = None):
         if queue_params:
-
-            cb = functools.partial(self.on_queue_declared,
+            cb = functools.partial(self.when_queue_declared,
                                    body,
                                    publish_params=publish_params,
                                    publish_key=publish_key)
-
-            self.declare_queue(queue_params, callback=cb)
-
+            self.declare_queue(queue_params, cb)
         else:
-            cb = functools.partial(self.on_exchange_declared,
+            cb = functools.partial(self.when_exchange_declared,
                                    body,
                                    exchange_params,
-                                   routing_key=routing_key,
+                                   routing_key,
                                    publish_params=publish_params,
                                    publish_key=publish_key)
+            self.declare_exchange(exchange_params, cb)
 
-            self.declare_exchange(exchange_params, callback=cb)
-
-    def on_queue_declared(self,
-                          body,
-                          frame,
-                          publish_params=None,
-                          publish_key=None):
-        """
-        :param body: bytes
-        :param frame: pika.frame.Method
-        :param publish_params: rabbitmq_client.PublishParams
-        :param publish_key: str
-        """
-        LOGGER.info(f"declared queue: {frame.method.queue}")
-
+    def when_queue_declared(self,
+                            body: bytes,
+                            queue_name: str,
+                            publish_params: PublishParams = None,
+                            publish_key: Union[None, str] = None):
         self._finalize_publish(body,
-                               routing_key=frame.method.queue,
+                               routing_key=queue_name,
                                publish_params=publish_params,
                                publish_key=publish_key)
 
-    def on_exchange_declared(self,
-                             body,
-                             exchange_params,
-                             _frame,
-                             routing_key="",
-                             publish_params=None,
-                             publish_key=None):
-        """
-        :param body: bytes
-        :param exchange_params: rabbitmq_client.ExchangeParams
-        :param _frame: pika.frame.Method
-        :param routing_key: str
-        :param publish_params: rabbitmq_client.PublishParams
-        :param publish_key: str
-        """
-        LOGGER.info(f"declared exchange: {exchange_params.exchange}")
-
+    def when_exchange_declared(self,
+                               body: bytes,
+                               exchange_params: ExchangeParams,
+                               routing_key: str,
+                               publish_params: PublishParams = None,
+                               publish_key: Union[None, str] = None):
         self._finalize_publish(body,
                                exchange=exchange_params.exchange,
                                routing_key=routing_key,
@@ -261,32 +231,25 @@ class RMQProducer(RMQConnection):
 
     def _finalize_publish(self,
                           body,
-                          exchange=DEFAULT_EXCHANGE,
-                          routing_key="",
-                          publish_params=None,
-                          publish_key=None):
-        """
-        :param body: bytes
-        :param exchange: str
-        :param routing_key: str
-        :param publish_params: rabbitmq_client.PublishParams
-        :param publish_key: str
-        """
-
+                          exchange: str = DEFAULT_EXCHANGE,
+                          routing_key: str = "",
+                          publish_params: PublishParams = None,
+                          publish_key: Union[None, str] = None):
         if publish_key is not None:
             with self._publish_lock:
+                LOGGER.debug(f"delivering with tag {self._next_delivery_tag}")
+
                 self._unacked_publishes[self._next_delivery_tag] = publish_key
                 self._next_delivery_tag += 1
 
                 self.basic_publish(body,
-                                   exchange=exchange,
-                                   routing_key=routing_key,
+                                   exchange,
+                                   routing_key,
                                    publish_params=publish_params)
-
         else:
             self.basic_publish(body,
-                               exchange=exchange,
-                               routing_key=routing_key,
+                               exchange,
+                               routing_key,
                                publish_params=publish_params)
 
     def _empty_buffered_messages(self):
@@ -294,12 +257,14 @@ class RMQProducer(RMQConnection):
         Sends all buffered messages.
         """
         for buffered_message in self._buffered_messages:
-            self._handle_publish(buffered_message.body,
-                                 buffered_message.exchange_params,
-                                 buffered_message.routing_key,
-                                 buffered_message.queue_params,
-                                 buffered_message.publish_params,
-                                 buffered_message.publish_key)
+            self._handle_publish(
+                buffered_message.body,
+                publish_params=buffered_message.publish_params,
+                queue_params=buffered_message.queue_params,
+                exchange_params=buffered_message.exchange_params,
+                routing_key=buffered_message.routing_key,
+                publish_key=buffered_message.publish_key
+            )
 
         self._buffered_messages = list()
 
@@ -317,19 +282,25 @@ class RMQProducer(RMQConnection):
         """
         :param frame: pika.frame.Method
         """
-        publish_key = self._unacked_publishes.pop(
-            frame.method.delivery_tag
-        )
+        LOGGER.debug(f"a delivery was confirmed: {frame.method.delivery_tag}")
 
-        if isinstance(frame.method, Basic.Ack):
-            self._confirm_delivery_callback(publish_key)
-
-        else:
-            LOGGER.error(f"broker nacked a publish: {frame}")
-            self._confirm_delivery_callback(
-                DeliveryError(publish_key)
+        try:
+            publish_key = self._unacked_publishes.pop(
+                frame.method.delivery_tag
             )
 
+            if isinstance(frame.method, Basic.Ack):
+                self._confirm_delivery_callback(publish_key)
+
+            else:
+                LOGGER.error(f"broker nacked a publish: {frame}")
+                self._confirm_delivery_callback(
+                    DeliveryError(publish_key)
+                )
+        except KeyError:
+            LOGGER.warning(f"RabbitMQ confirmed unexpected delivery tag:"
+                           f" {frame.method.delivery_tag}")
+
     def on_ready(self):
         """
         Connection hook, called when channel opened, meaning RMQConnection is
@@ -357,7 +328,7 @@ class RMQProducer(RMQConnection):
 
         :param permanent: bool
         """
-        if permanent:
+        if not self._closing and permanent:
             LOGGER.critical("producer connection permanently closed")
         else:
             LOGGER.info("producer connection closed")
@@ -368,15 +339,19 @@ class RMQProducer(RMQConnection):
         # the tag is reset.
         self._next_delivery_tag = 1
 
-    def on_error(self):
+    def on_error(self, error: Union[MandatoryError, DeclarationError]):
         """
         Connection hook, called when the connection has encountered an error.
         """
         LOGGER.info("producer connection error")
 
-        raise NotImplementedError
-        # Possible errors:
-        # * channel died and will not recover
-        # * callback for operation failed
-        # * declaration with faulty parameters attempted
-        # TODO: Add possibility to signal user that an error has occurred.
+        if isinstance(error, MandatoryError):
+            if self._confirm_delivery_callback is not None:
+                self._confirm_delivery_callback(error)
+            else:
+                LOGGER.error(f"failed to publish to "
+                             f"'{error.exchange}' + '{error.routing_key}', "
+                             f"no queue is bound to it")
+
+        elif isinstance(error, DeclarationError):
+            LOGGER.error(f"failed to declare something: {error.message}")
